#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import io
import sys
import logging as log

#~~ version

from ._version import get_versions
versions = get_versions()

__version__ = versions['version']
__branch__ = versions.get('branch', None)
__display_version__ = __version__
__revision__ = versions.get('full-revisionid', versions.get('full', None))

del versions
del get_versions

#~~ try to ensure a sound SSL environment

urllib3_ssl = True
"""Whether requests/urllib3 and urllib3 (if installed) should be able to establish
   a sound SSL environment or not."""

version_info = sys.version_info
if version_info.major == 2 and version_info.minor <= 7 and version_info.micro < 9:
	try:
		# make sure our requests version of urllib3 is properly patched (if possible)
		import requests.packages.urllib3.contrib.pyopenssl
		requests.packages.urllib3.contrib.pyopenssl.inject_into_urllib3()
	except ImportError:
		urllib3_ssl = False

	try:
		import urllib3

		# only proceed if urllib3 is even installed on its own
		try:
			# urllib3 is there, let's patch that too
			import urllib3.contrib.pyopenssl
			urllib3.contrib.pyopenssl.inject_into_urllib3()
		except ImportError:
			urllib3_ssl = False
	except ImportError:
		pass

del version_info

#~~ custom exceptions

class FatalStartupError(Exception):
	def __init__(self, message, cause=None):
		self.cause = cause
		Exception.__init__(self, message)

	def __str__(self):
		result = Exception.__str__(self)
		if self.cause:
			return "{}: {}".format(result, str(self.cause))
		else:
			return result

#~~ init methods to bring up platform

def init_platform(basedir, configfile, use_logging_file=True, logging_file=None,
                  logging_config=None, debug=False, verbosity=0, uncaught_logger=None,
                  uncaught_handler=None, safe_mode=False, ignore_blacklist=False, after_preinit_logging=None,
                  after_settings_init=None, after_logging=None, after_safe_mode=None, after_settings_valid=None,
                  after_event_manager=None, after_connectivity_checker=None,
                  after_plugin_manager=None, after_environment_detector=None):
	kwargs = dict()

	logger, recorder = preinit_logging(debug, verbosity, uncaught_logger, uncaught_handler)
	kwargs["logger"] = logger
	kwargs["recorder"] = recorder

	if callable(after_preinit_logging):
		after_preinit_logging(**kwargs)

	try:
		settings = init_settings(basedir, configfile)
	except Exception as ex:
		raise FatalStartupError("Could not initialize settings manager", cause=ex)
	kwargs["settings"] = settings
	if callable(after_settings_init):
		after_settings_init(**kwargs)

	try:
		logger = init_logging(settings,
		                      use_logging_file=use_logging_file,
		                      logging_file=logging_file,
		                      default_config=logging_config,
		                      debug=debug,
		                      verbosity=verbosity,
		                      uncaught_logger=uncaught_logger,
		                      uncaught_handler=uncaught_handler)
	except Exception as ex:
		raise FatalStartupError("Could not initialize logging", cause=ex)

	kwargs["logger"] = logger
	if callable(after_logging):
		after_logging(**kwargs)

	settings_start_once_in_safemode = "settings" if settings.getBoolean(["server", "startOnceInSafeMode"]) else None
	settings_incomplete_startup_safemode = "incomplete_startup" \
		if settings.getBoolean(["server", "incompleteStartup"]) \
		   and not settings.getBoolean(["server", "ignoreIncompleteStartup"]) \
		else None
	safe_mode = safe_mode or settings_start_once_in_safemode or settings_incomplete_startup_safemode
	kwargs["safe_mode"] = safe_mode
	if callable(after_safe_mode):
		after_safe_mode(**kwargs)

	# now before we continue, let's make sure *all* our folders are sane
	try:
		settings.sanity_check_folders()
	except Exception as ex:
		raise FatalStartupError("Configured folders didn't pass sanity check", cause=ex)
	if callable(after_settings_valid):
		after_settings_valid(**kwargs)

	try:
		event_manager = init_event_manager(settings)
	except Exception as ex:
		raise FatalStartupError("Could not initialize event manager", cause=ex)

	kwargs["event_manager"] = event_manager
	if callable(after_event_manager):
		after_event_manager(**kwargs)

	try:
		connectivity_checker = init_connectivity_checker(settings, event_manager)
	except Exception as ex:
		raise FatalStartupError("Could not initialize connectivity checker", cause=ex)

	kwargs["connectivity_checker"] = connectivity_checker
	if callable(after_connectivity_checker):
		after_connectivity_checker(**kwargs)

	try:
		plugin_manager = init_pluginsystem(settings,
		                                   safe_mode=safe_mode,
		                                   ignore_blacklist=ignore_blacklist,
		                                   connectivity_checker=connectivity_checker)
	except Exception as ex:
		raise FatalStartupError("Could not initialize plugin manager", cause=ex)

	kwargs["plugin_manager"] = plugin_manager
	if callable(after_plugin_manager):
		after_plugin_manager(**kwargs)

	try:
		environment_detector = init_environment_detector(plugin_manager)
	except Exception as ex:
		raise FatalStartupError("Could not initialize environment detector", cause=ex)

	kwargs["environment_detector"] = environment_detector
	if callable(after_environment_detector):
		after_environment_detector(**kwargs)

	return settings, logger, safe_mode, event_manager, connectivity_checker, plugin_manager, environment_detector


def init_settings(basedir, configfile):
	"""Inits the settings instance based on basedir and configfile to use."""

	from octoprint.settings import settings, InvalidSettings
	try:
		return settings(init=True, basedir=basedir, configfile=configfile)
	except InvalidSettings as e:
		message = "Error parsing the configuration file, it appears to be invalid YAML."
		if e.line is not None and e.column is not None:
			message += " The parser reported an error on line {}, column {}.".format(e.line, e.column)
		raise FatalStartupError(message)


def preinit_logging(debug=False, verbosity=0, uncaught_logger=None, uncaught_handler=None):
	config = {
		"version": 1,
		"formatters": {
			"simple": {
				"format": "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
			}
		},
		"handlers": {
			"console": {
				"class": "logging.StreamHandler",
				"level": "DEBUG",
				"formatter": "simple",
				"stream": "ext://sys.stdout"
			}
		},
		"loggers": {
			"octoprint": {
				"level": "DEBUG" if debug else "INFO"
			},
			"octoprint.util": {
				"level": "INFO"
			}
		},
		"root": {
			"level": "WARN",
			"handlers": ["console"]
		}
	}

	logger = set_logging_config(config, debug, verbosity, uncaught_logger, uncaught_handler)

	from octoprint.logging.handlers import RecordingLogHandler
	recorder = RecordingLogHandler(level=log.DEBUG)
	log.getLogger().addHandler(recorder)

	return logger, recorder


def init_logging(settings, use_logging_file=True, logging_file=None, default_config=None, debug=False, verbosity=0, uncaught_logger=None, uncaught_handler=None):
	"""Sets up logging."""

	import os

	from octoprint.util import dict_merge

	# default logging configuration
	if default_config is None:
		default_config = {
			"version": 1,
			"formatters": {
				"simple": {
					"format": "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
				},
				"serial": {
					"format": "%(asctime)s - %(message)s"
				}
			},
			"handlers": {
				"console": {
					"class": "octoprint.logging.handlers.OctoPrintStreamHandler",
					"level": "DEBUG",
					"formatter": "simple",
					"stream": "ext://sys.stdout"
				},
				"file": {
					"class": "octoprint.logging.handlers.OctoPrintLogHandler",
					"level": "DEBUG",
					"formatter": "simple",
					"when": "D",
					"backupCount": 6,
					"filename": os.path.join(settings.getBaseFolder("logs"), "octoprint.log")
				},
				"serialFile": {
					"class": "octoprint.logging.handlers.SerialLogHandler",
					"level": "DEBUG",
					"formatter": "serial",
					"backupCount": 3,
					"filename": os.path.join(settings.getBaseFolder("logs"), "serial.log"),
					"delay": True
				}
			},
			"loggers": {
				"SERIAL": {
					"level": "INFO",
					"handlers": ["serialFile"],
					"propagate": False
				},
				"octoprint": {
					"level": "INFO"
				},
				"octoprint.util": {
					"level": "INFO"
				},
				"octoprint.plugins": {
					"level": "INFO"
				}
			},
			"root": {
				"level": "WARN",
				"handlers": ["console", "file"]
			}
		}

	if debug or verbosity > 0:
		default_config["loggers"]["octoprint"]["level"] = "DEBUG"
		default_config["root"]["level"] = "INFO"
	if verbosity > 1:
		default_config["loggers"]["octoprint.plugins"]["level"] = "DEBUG"
	if verbosity > 2:
		default_config["root"]["level"] = "DEBUG"

	config = default_config
	if use_logging_file:
		# further logging configuration from file...
		if logging_file is None:
			logging_file = os.path.join(settings.getBaseFolder("base"), "logging.yaml")

		config_from_file = {}
		if os.path.exists(logging_file) and os.path.isfile(logging_file):
			import yaml
			with io.open(logging_file, 'rt', encoding='utf-8') as f:
				config_from_file = yaml.safe_load(f)

		# we merge that with the default config
		if config_from_file is not None and isinstance(config_from_file, dict):
			config = dict_merge(default_config, config_from_file)

	# configure logging globally
	return set_logging_config(config, debug, verbosity, uncaught_logger, uncaught_handler)


def set_logging_config(config, debug, verbosity, uncaught_logger, uncaught_handler):
	# configure logging globally
	import logging.config as logconfig
	logconfig.dictConfig(config)

	# make sure we log any warnings
	log.captureWarnings(True)

	import warnings

	categories = (DeprecationWarning, PendingDeprecationWarning)
	if verbosity > 2:
		warnings.simplefilter("always")
	elif debug or verbosity > 0:
		for category in categories:
			warnings.simplefilter("always", category=category)

	# make sure we also log any uncaught exceptions
	if uncaught_logger is None:
		logger = log.getLogger(__name__)
	else:
		logger = log.getLogger(uncaught_logger)

	if uncaught_handler is None:
		def exception_logger(exc_type, exc_value, exc_tb):
			logger.error("Uncaught exception", exc_info=(exc_type, exc_value, exc_tb))

		uncaught_handler = exception_logger
	sys.excepthook = uncaught_handler

	return logger


def init_pluginsystem(settings, safe_mode=False, ignore_blacklist=True, connectivity_checker=None):
	"""Initializes the plugin manager based on the settings."""

	import os

	# we need this so that octoprint.plugins is in sys.modules and no warnings are caused when loading bundled plugins
	import octoprint.plugins

	logger = log.getLogger(__name__ + ".startup")

	plugin_folders = [(os.path.abspath(os.path.join(os.path.dirname(os.path.realpath(__file__)), "plugins")), "octoprint.plugins", True),
	                  settings.getBaseFolder("plugins")]
	plugin_entry_points = ["octoprint.plugin"]
	plugin_disabled_list = settings.get(["plugins", "_disabled"])

	plugin_blacklist = []
	if not ignore_blacklist and settings.getBoolean(["server", "pluginBlacklist", "enabled"]):
		plugin_blacklist = get_plugin_blacklist(settings, connectivity_checker=connectivity_checker)

	plugin_validators = []
	if safe_mode:
		def validator(phase, plugin_info):
			if phase in ("before_import", "before_load", "before_enable"):
				setattr(plugin_info, "safe_mode_victim", not plugin_info.bundled)
				if not plugin_info.bundled:
					return False
			return True
		plugin_validators.append(validator)

	compatibility_ignored_list = settings.get(["plugins", "_forcedCompatible"])

	from octoprint.plugin import plugin_manager
	pm = plugin_manager(init=True,
	                    plugin_folders=plugin_folders,
	                    plugin_entry_points=plugin_entry_points,
	                    plugin_disabled_list=plugin_disabled_list,
	                    plugin_blacklist=plugin_blacklist,
	                    plugin_validators=plugin_validators,
	                    compatibility_ignored_list=compatibility_ignored_list)

	settings_overlays = dict()
	disabled_from_overlays = dict()

	def handle_plugin_loaded(name, plugin):
		if plugin.instance and hasattr(plugin.instance, "__plugin_settings_overlay__"):
			plugin.needs_restart = True

			# plugin has a settings overlay, inject it
			overlay_definition = getattr(plugin.instance, "__plugin_settings_overlay__")
			if isinstance(overlay_definition, (tuple, list)):
				overlay_definition, order = overlay_definition
			else:
				order = None

			overlay = settings.load_overlay(overlay_definition)

			if "plugins" in overlay and "_disabled" in overlay["plugins"]:
				disabled_plugins = overlay["plugins"]["_disabled"]
				del overlay["plugins"]["_disabled"]
				disabled_from_overlays[name] = (disabled_plugins, order)

			settings_overlays[name] = overlay
			logger.debug("Found settings overlay on plugin {}".format(name))

	def handle_plugins_loaded(startup=False, initialize_implementations=True, force_reload=None):
		if not startup:
			return

		from octoprint.util import sv

		sorted_disabled_from_overlays = sorted([(key, value[0], value[1]) for key, value in disabled_from_overlays.items()], key=lambda x: (x[2] is None, sv(x[2]), sv(x[0])))

		disabled_list = pm.plugin_disabled_list
		already_processed = []
		for name, addons, _ in sorted_disabled_from_overlays:
			if not name in disabled_list and not name.endswith("disabled"):
				for addon in addons:
					if addon in disabled_list:
						continue

					if addon in already_processed:
						logger.info("Plugin {} wants to disable plugin {}, but that was already processed".format(name, addon))

					if not addon in already_processed and not addon in disabled_list:
						disabled_list.append(addon)
						logger.info("Disabling plugin {} as defined by plugin {}".format(addon, name))
				already_processed.append(name)

	def handle_plugin_enabled(name, plugin):
		if name in settings_overlays:
			settings.add_overlay(settings_overlays[name])
			logger.info("Added settings overlay from plugin {}".format(name))

	pm.on_plugin_loaded = handle_plugin_loaded
	pm.on_plugins_loaded = handle_plugins_loaded
	pm.on_plugin_enabled = handle_plugin_enabled
	pm.reload_plugins(startup=True, initialize_implementations=False)
	return pm


def get_plugin_blacklist(settings, connectivity_checker=None):
	import requests
	import os
	import time
	import yaml

	from octoprint.util import bom_aware_open
	from octoprint.util.version import is_octoprint_compatible

	logger = log.getLogger(__name__ + ".startup")

	if connectivity_checker is not None and not connectivity_checker.online:
		logger.info("We don't appear to be online, not fetching plugin blacklist")
		return []

	def format_blacklist(entries):
		format_entry = lambda x: "{} ({})".format(x[0], x[1]) if isinstance(x, (list, tuple)) and len(x) == 2 \
			else "{} (any)".format(x)
		return ", ".join(map(format_entry, entries))

	def process_blacklist(entries):
		result = []

		if not isinstance(entries, list):
			return result

		for entry in entries:
			if not "plugin" in entry:
				continue

			if "octoversions" in entry and not is_octoprint_compatible(*entry["octoversions"]):
				continue

			if "version" in entry:
				logger.debug("Blacklisted plugin: {}, version: {}".format(entry["plugin"], entry["version"]))
				result.append((entry["plugin"], entry["version"]))
			elif "versions" in entry:
				logger.debug("Blacklisted plugin: {}, versions: {}".format(entry["plugin"], ", ".join(entry["versions"])))
				for version in entry["versions"]:
					result.append((entry["plugin"], version))
			else:
				logger.debug("Blacklisted plugin: {}".format(entry["plugin"]))
				result.append(entry["plugin"])

		return result

	def fetch_blacklist_from_cache(path, ttl):
		if not os.path.isfile(path):
			return None

		if os.stat(path).st_mtime + ttl < time.time():
			return None

		with bom_aware_open(path, encoding="utf-8", mode='rt') as f:
			result = yaml.safe_load(f)

		if isinstance(result, list):
			return result

	def fetch_blacklist_from_url(url, timeout=3, cache=None):
		result = []
		try:
			r = requests.get(url, timeout=timeout)
			result = process_blacklist(r.json())

			if cache is not None:
				try:
					with bom_aware_open(cache, encoding="utf-8", mode='wt') as f:
						yaml.safe_dump(result, f)
				except Exception as e:
					logger.info("Fetched plugin blacklist but couldn't write it to its cache file: %s", e)
		except Exception as e:
			logger.info("Unable to fetch plugin blacklist from {}, proceeding without it: {}".format(url, e))
		return result

	try:
		# first attempt to fetch from cache
		cache_path = os.path.join(settings.getBaseFolder("data"), "plugin_blacklist.yaml")
		ttl = settings.getInt(["server", "pluginBlacklist", "ttl"])
		blacklist = fetch_blacklist_from_cache(cache_path, ttl)

		if blacklist is None:
			# no result from the cache, let's fetch it fresh
			url = settings.get(["server", "pluginBlacklist", "url"])
			timeout = settings.getFloat(["server", "pluginBlacklist", "timeout"])
			blacklist = fetch_blacklist_from_url(url, timeout=timeout, cache=cache_path)

		if blacklist is None:
			# still now result, so no blacklist
			blacklist = []

		if blacklist:
			logger.info("Blacklist processing done, "
			            "adding {} blacklisted plugin versions: {}".format(len(blacklist),
			                                                               format_blacklist(blacklist)))
		else:
			logger.info("Blacklist processing done")

		return blacklist
	except Exception:
		logger.exception("Something went wrong while processing the plugin blacklist. Proceeding without it.")


def init_event_manager(settings):
	from octoprint.events import eventManager
	return eventManager()


def init_connectivity_checker(settings, event_manager):
	from octoprint.events import Events
	from octoprint.util import ConnectivityChecker

	# start regular check if we are connected to the internet
	connectivityEnabled = settings.getBoolean(["server", "onlineCheck", "enabled"])
	connectivityInterval = settings.getInt(["server", "onlineCheck", "interval"])
	connectivityHost = settings.get(["server", "onlineCheck", "host"])
	connectivityPort = settings.getInt(["server", "onlineCheck", "port"])

	def on_connectivity_change(old_value, new_value):
		event_manager.fire(Events.CONNECTIVITY_CHANGED, payload=dict(old=old_value, new=new_value))

	connectivityChecker = ConnectivityChecker(connectivityInterval,
	                                          connectivityHost,
	                                          port=connectivityPort,
	                                          enabled=connectivityEnabled,
	                                          on_change=on_connectivity_change)

	return connectivityChecker


def init_environment_detector(plugin_manager):
	from octoprint.environment import EnvironmentDetector
	return EnvironmentDetector(plugin_manager)

#~~ server main method

def main():
	import sys

	# os args are gained differently on win32
	try:
		from click.utils import get_os_args
		args = get_os_args()
	except ImportError:
		# for whatever reason we are running an older Click version?
		args = sys.argv[1:]

	if len(args) >= len(sys.argv):
		# Now some ugly preprocessing of our arguments starts. We have a somewhat difficult situation on our hands
		# here if we are running under Windows and want to be able to handle utf-8 command line parameters (think
		# plugin parameters such as names or something, e.g. for the "dev plugin:new" command) while at the same
		# time also supporting sys.argv rewriting for debuggers etc (e.g. PyCharm).
		#
		# So what we try to do here is solve this... Generally speaking, sys.argv and whatever Windows returns
		# for its CommandLineToArgvW win32 function should have the same length. If it doesn't however and
		# sys.argv is shorter than the win32 specific command line arguments, obviously stuff was cut off from
		# sys.argv which also needs to be cut off of the win32 command line arguments.
		#
		# So this is what we do here.

		# -1 because first entry is the script that was called
		sys_args_length = len(sys.argv) - 1

		# cut off stuff from the beginning
		args = args[-1 * sys_args_length:] if sys_args_length else []

	from octoprint.util.fixes import patch_sarge_async_on_py2
	patch_sarge_async_on_py2()

	from octoprint.cli import octo
	octo(args=args, prog_name="octoprint", auto_envvar_prefix="OCTOPRINT")


if __name__ == "__main__":
	main()



# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2019 The OctoPrint Project - Released under terms of the AGPLv3 License"

import logging
import pkg_resources
import requests

from octoprint.util import to_native_str
from octoprint.util.version import is_python_compatible, get_comparable_version, is_prerelease

INFO_URL = "https://pypi.org/pypi/{package}/json"

logger = logging.getLogger("octoprint.plugins.softwareupdate.version_checks.pypi_release")

def _filter_out_latest(releases, include_prerelease=False, python_version=None):
	"""
	Filters out the newest of all matching releases.

	Tests:

	    >>> requires_py2 = ">=2.7.9,<3"
	    >>> requires_py23 = ">=2.7.9, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*, <4"
	    >>> requires_py3 = ">=3.6, <4"
	    >>> releases = {"1.3.12": [dict(requires_python=requires_py2, upload_time_iso_8601="2019-10-22T10:06:03.190293Z")], "1.4.0rc1": [dict(requires_python=requires_py23, upload_time_iso_8601="2019-11-22T10:06:03.190293Z")], "2.0.0rc1": [dict(requires_python=requires_py3, upload_time_iso_8601="2020-10-22T10:06:03.190293Z")]}
	    >>> to_native_str(_filter_out_latest(releases, python_version="2.7.9"))
	    '1.3.12'
	    >>> to_native_str(_filter_out_latest(releases, include_prerelease=True, python_version="2.7.9"))
	    '1.4.0rc1'
	    >>> to_native_str(_filter_out_latest(releases, include_prerelease=True, python_version="3.6.0"))
	    '2.0.0rc1'
	    >>> to_native_str(_filter_out_latest(releases, python_version="3.6.0"))
	"""
	releases = [dict(version=k, data=v[0]) for k, v in releases.items()]

	# filter out prereleases and versions incompatible to our python
	filter_function = lambda release: not is_prerelease(release["version"]) and is_python_compatible(release["data"].get("requires_python", ""), python_version=python_version)
	if include_prerelease:
		filter_function = lambda release: is_python_compatible(release["data"].get("requires_python", ""), python_version=python_version)

	releases = list(filter(filter_function, releases))
	if not releases:
		return None

	# sort by upload date
	releases = sorted(releases, key=lambda release: release["data"].get("upload_time_iso_8601", ""))

	# latest release = last in list
	latest = releases[-1]

	return latest["version"]

def _get_latest_release(package, include_prerelease):
	from ..exceptions import NetworkError

	try:
		r = requests.get(INFO_URL.format(package=package), timeout=(3.05, 30))
	except requests.ConnectionError as exc:
		raise NetworkError(cause=exc)

	if not r.status_code == requests.codes.ok:
		return None

	data = r.json()
	if not "info" in data or not "version" in data["info"]:
		return None

	requires_python = data["info"].get("requires_python")
	if requires_python and not is_python_compatible(requires_python):
		return None

	return _filter_out_latest(data["releases"],
	                          include_prerelease=include_prerelease)

def _is_current(release_information):
	if release_information["remote"]["value"] is None:
		return True

	local_version = get_comparable_version(release_information["local"]["value"])
	remote_version = get_comparable_version(release_information["remote"]["value"])

	return remote_version <= local_version

def get_latest(target, check, online=True):
	from ..exceptions import CannotUpdateOffline

	if not online and not check.get("offline", False):
		raise CannotUpdateOffline()

	package = check.get("package")

	distribution = pkg_resources.get_distribution(package)
	if distribution:
		local_version = distribution.version
	else:
		local_version = None

	remote_version = _get_latest_release(package,
	                                     include_prerelease=check.get("prerelease", False))

	information = dict(local=dict(name=local_version, value=local_version),
	                   remote=dict(name=remote_version, value=remote_version))

	logger.debug("Target: {}, local: {}, remote: {}".format(target, information["local"]["name"], information["remote"]["name"]))

	return information, _is_current(information)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

"""
This module represents OctoPrint's settings management. Within this module the default settings for the core
application are defined and the instance of the :class:`Settings` is held, which offers getter and setter
methods for the raw configuration values as well as various convenience methods to access the paths to base folders
of various types and the configuration file itself.

.. autodata:: default_settings
   :annotation: = dict(...)

.. autodata:: valid_boolean_trues

.. autofunction:: settings

.. autoclass:: Settings
   :members:
   :undoc-members:
"""

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

import io
import sys
import os
import yaml
import yaml.parser
import logging
import re
import uuid
import copy
import time

# noinspection PyCompatibility
from builtins import bytes
# noinspection PyCompatibility
from past.builtins import basestring

try:
	from collections import ChainMap
except ImportError:
	from chainmap import ChainMap

try:
	from collections.abc import KeysView
except ImportError:
	from collections import KeysView

from octoprint.util import atomic_write, is_hidden_path, dict_merge, CaseInsensitiveSet, generate_api_key

_APPNAME = "OctoPrint"

_instance = None

def settings(init=False, basedir=None, configfile=None):
	"""
	Factory method for initially constructing and consecutively retrieving the :class:`~octoprint.settings.Settings`
	singleton.

	Arguments:
	    init (boolean): A flag indicating whether this is the initial call to construct the singleton (True) or not
	        (False, default). If this is set to True and the plugin manager has already been initialized, a :class:`ValueError`
	        will be raised. The same will happen if the plugin manager has not yet been initialized and this is set to
	        False.
	    basedir (str): Path of the base directory for all of OctoPrint's settings, log files, uploads etc. If not set
	        the default will be used: ``~/.octoprint`` on Linux, ``%APPDATA%/OctoPrint`` on Windows and
	        ``~/Library/Application Support/OctoPrint`` on MacOS.
	    configfile (str): Path of the configuration file (``config.yaml``) to work on. If not set the default will
	        be used: ``<basedir>/config.yaml`` for ``basedir`` as defined above.

	Returns:
	    Settings: The fully initialized :class:`Settings` instance.

	Raises:
	    ValueError: ``init`` is True but settings are already initialized or vice versa.
	"""
	global _instance
	if _instance is not None:
		if init:
			raise ValueError("Settings Manager already initialized")

	else:
		if init:
			_instance = Settings(configfile=configfile, basedir=basedir)
		else:
			raise ValueError("Settings not initialized yet")

	return _instance

default_settings = {
	"serial": {
		"port": None,
		"baudrate": None,
		"exclusive": True,
		"autoconnect": False,
		"log": False,
		"timeout": {
			"detection": 1,
			"connection": 10,
			"communication": 30,
			"communicationBusy": 3,
			"temperature": 5,
			"temperatureTargetSet": 2,
			"temperatureAutoreport": 2,
			"sdStatus": 1,
			"sdStatusAutoreport": 1,
			"resendOk": .5,
			"baudrateDetectionPause": 1.0,
			"positionLogWait": 10.0,
		},
		"maxCommunicationTimeouts": {
			"idle": 2,
			"printing": 5,
			"long": 5
		},
		"maxWritePasses": 5,
		"additionalPorts": [],
		"additionalBaudrates": [],
		"longRunningCommands": ["G4", "G28", "G29", "G30", "G32", "M400", "M226", "M600"],
		"blockedCommands": ["M0", "M1"],
		"pausingCommands": ["M0", "M1", "M25"],
		"emergencyCommands": ["M112", "M108", "M410"],
		"checksumRequiringCommands": ["M110"],
		"helloCommand": "M110 N0",
		"disconnectOnErrors": True,
		"ignoreErrorsFromFirmware": False,
		"terminalLogSize": 20,
		"logResends": True,
		"supportResendsWithoutOk": "detect",
		"logPositionOnPause": True,
		"logPositionOnCancel": False,
		"abortHeatupOnCancel": True,
		"waitForStartOnConnect": False,
		"alwaysSendChecksum": False,
		"neverSendChecksum": False,
		"sendChecksumWithUnknownCommands": False,
		"unknownCommandsNeedAck": False,
		"sdRelativePath": False,
		"sdAlwaysAvailable": False,
		"maxNotSdPrinting": 2,
		"swallowOkAfterResend": True,
		"repetierTargetTemp": False,
		"externalHeatupDetection": True,
		"supportWait": True,
		"ignoreIdenticalResends": False,
		"identicalResendsCountdown": 7,
		"supportFAsCommand": False,
		"firmwareDetection": True,
		"blockWhileDwelling": False,
		"useParityWorkaround": "detect",
		"maxConsecutiveResends": 10,
		"sendM112OnError": True,
		"ackMax": 1,

		"capabilities": {
			"autoreport_temp": True,
			"autoreport_sdstatus": True,
			"busy_protocol": True,
			"emergency_parser": True
		},

		# command specific flags
		"triggerOkForM29": True
	},
	"server": {
		"host": None,
		"port": 5000,
		"firstRun": True,
		"startOnceInSafeMode": False,
		"ignoreIncompleteStartup": False,
		"incompleteStartup": False,
		"seenWizards": {},
		"secretKey": None,
		"heartbeat": 15 * 60, # 15 min
		"reverseProxy": {
			"prefixHeader": None,
			"schemeHeader": None,
			"hostHeader": None,
			"serverHeader": None,
			"portHeader": None,
			"prefixFallback": None,
			"schemeFallback": None,
			"hostFallback": None,
			"serverFallback": None,
			"portFallback": None,
			"trustedDownstream": []
		},
		"uploads": {
			"maxSize":  1 * 1024 * 1024 * 1024, # 1GB
			"nameSuffix": "name",
			"pathSuffix": "path"
		},
		"maxSize": 100 * 1024, # 100 KB
		"commands": {
			"systemShutdownCommand": None,
			"systemRestartCommand": None,
			"serverRestartCommand": None
		},
		"onlineCheck": {
			"enabled": None,
			"interval": 15 * 60, # 15 min
			"host": "8.8.8.8",
			"port": 53
		},
		"pluginBlacklist": {
			"enabled": None,
			"url": "https://plugins.octoprint.org/blacklist.json",
			"ttl": 15 * 60 # 15 min
		},
		"diskspace": {
			"warning": 500 * 1024 * 1024, # 500 MB
			"critical": 200 * 1024 * 1024, # 200 MB
		},
		"preemptiveCache": {
			"exceptions": [],
			"until": 7
		},
		"ipCheck": {
			"enabled": True,
			"trustedSubnets": []
		},
		"allowFraming": False
	},
	"webcam": {
		"webcamEnabled": True,
		"timelapseEnabled": True,
		"stream": None,
		"streamRatio": "16:9",
		"streamTimeout": 5,
		"snapshot": None,
		"snapshotTimeout": 5,
		"snapshotSslValidation": True,
		"ffmpeg": None,
		"ffmpegThreads": 1,
		"ffmpegVideoCodec": "mpeg2video",
		"bitrate": "10000k",
		"watermark": True,
		"flipH": False,
		"flipV": False,
		"rotate90" : False,
		"timelapse": {
			"type": "off",
			"options": {},
			"postRoll": 0,
			"fps": 25,
		},
		"cleanTmpAfterDays": 7
	},
	"gcodeViewer": {
		"enabled": True,
		"mobileSizeThreshold": 2 * 1024 * 1024, # 2MB
		"sizeThreshold": 20 * 1024 * 1024, # 20MB
	},
	"gcodeAnalysis": {
		"maxExtruders": 10,
		"throttle_normalprio": 0.01,
		"throttle_highprio": 0.0,
		"throttle_lines": 100,
		"runAt": "idle" # 'never', 'idle', 'always'
	},
	"feature": {
		"temperatureGraph": True,
		"sdSupport": True,
		"keyboardControl": True,
		"pollWatched": False,
		"modelSizeDetection": True,
		"printStartConfirmation": False,
		"printCancelConfirmation": True,
		"autoUppercaseBlacklist": ["M117", "M118"],
		"g90InfluencesExtruder": False
	},
	"folder": {
		"uploads": None,
		"timelapse": None,
		"timelapse_tmp": None,
		"logs": None,
		"virtualSd": None,
		"watched": None,
		"plugins": None,
		"slicingProfiles": None,
		"printerProfiles": None,
		"scripts": None,
		"translations": None,
		"generated": None,
		"data": None
	},
	"temperature": {
		"profiles": [
			{"name": "ABS", "extruder" : 210, "bed" : 100 },
			{"name": "PLA", "extruder" : 180, "bed" : 60 }
		],
		"cutoff": 30,
		"sendAutomatically": False,
		"sendAutomaticallyAfter": 1,
	},
	"printerProfiles": {
		"default": None
	},
	"printerParameters": {
		"pauseTriggers": [],
		"defaultExtrusionLength": 5
	},
	"appearance": {
		"name": "",
		"color": "default",
		"colorTransparent": False,
		"colorIcon": True,
		"defaultLanguage": "_default",
		"showFahrenheitAlso": False,
		"fuzzyTimes": True,
		"closeModalsWithClick": True,
		"components": {
			"order": {
				"navbar": ["settings", "systemmenu", "plugin_announcements", "plugin_pi_support", "login"],
				"sidebar": ["plugin_printer_safety_check", "connection", "state", "files"],
				"tab": ["temperature", "control", "gcodeviewer", "terminal", "timelapse"],
				"settings": [
					"section_printer", "serial", "printerprofiles", "temperatures", "terminalfilters", "gcodescripts",
					"section_features", "features", "webcam", "accesscontrol", "gcodevisualizer", "api", "plugin_appkeys",
					"section_octoprint", "server", "folders", "appearance", "plugin_logging", "plugin_pluginmanager",
					"plugin_softwareupdate", "plugin_announcements", "plugin_backup", "plugin_tracking", "plugin_errortracking",
					"plugin_pi_support"
				],
				"usersettings": ["access", "interface"],
				"wizard": ["access"],
				"about": ["about", "plugin_pi_support", "supporters", "authors", "changelog", "license", "thirdparty", "plugin_pluginmanager"],
				"generic": []
			},
			"disabled": {
				"navbar": [],
				"sidebar": [],
				"tab": [],
				"settings": [],
				"usersettings": [],
				"generic": []
			}
		}
	},
	"controls": [],
	"system": {
		"actions": []
	},
	"accessControl": {
		"enabled": True,
		"salt": None,
		"userManager": "octoprint.access.users.FilebasedUserManager",
		"groupManager": "octoprint.access.groups.FilebasedGroupManager",
		"permissionManager": "octoprint.access.permissions.PermissionManager",
		"userfile": None,
		"groupfile": None,
		"autologinLocal": False,
		"localNetworks": ["127.0.0.0/8", "::1/128"],
		"autologinAs": None,
		"trustBasicAuthentication": False,
		"checkBasicAuthenticationPassword": True,
		"trustRemoteUser": False,
		"remoteUserHeader": "REMOTE_USER",
		"addRemoteUsers": False
	},
	"slicing": {
		"enabled": True,
		"defaultSlicer": None,
		"defaultProfiles": None
	},
	"events": {
		"enabled": True,
		"subscriptions": []
	},
	"api": {
		"key": None,
		"allowCrossOrigin": False,
		"apps": {}
	},
	"terminalFilters": [
		{ "name": "Suppress temperature messages", "regex": r"(Send: (N\d+\s+)?M105)|(Recv:\s+(ok\s+((P|B|N)\d+\s+)*)?(B|T\d*):\d+)" },
		{ "name": "Suppress SD status messages", "regex": r"(Send: (N\d+\s+)?M27)|(Recv: SD printing byte)|(Recv: Not SD printing)" },
		{ "name": "Suppress wait responses", "regex": "Recv: wait"}
	],
	"plugins": {
		"_disabled": [],
		"_forcedCompatible": []
	},
	"scripts": {
		"gcode": {
			"afterPrintCancelled": "; disable motors\nM84\n\n;disable all heaters\n{% snippet 'disable_hotends' %}\n{% snippet 'disable_bed' %}\n;disable fan\nM106 S0",
			"snippets": {
				"disable_hotends": "{% if printer_profile.extruder.sharedNozzle %}M104 T0 S0\n{% else %}{% for tool in range(printer_profile.extruder.count) %}M104 T{{ tool }} S0\n{% endfor %}{% endif %}",
				"disable_bed": "{% if printer_profile.heatedBed %}M140 S0\n{% endif %}"
			}
		}
	},
	"estimation": {
		"printTime": {
			"statsWeighingUntil": 0.5,
			"validityRange": 0.15,
			"forceDumbFromPercent": 0.3,
			"forceDumbAfterMin": 30,
			"stableThreshold": 60
		}
	},
	"devel": {
		"stylesheet": "css",
		"cache": {
			"enabled": True,
			"preemptive": True
		},
		"webassets": {
			"bundle": True,
			"clean_on_startup": True
		},
		"useFrozenDictForPrinterState": True,
		"showLoadingAnimation": True,
		"virtualPrinter": {
			"enabled": False,
			"okAfterResend": False,
			"forceChecksum": False,
			"numExtruders": 1,
			"pinnedExtruders": None,
			"includeCurrentToolInTemps": True,
			"includeFilenameInOpened": True,
			"hasBed": True,
			"hasChamber": False,
			"repetierStyleTargetTemperature": False,
			"okBeforeCommandOutput": False,
			"smoothieTemperatureReporting": False,
			"reprapfwM114": False,
			"extendedSdFileList": False,
			"throttle": 0.01,
			"sendWait": True,
			"waitInterval": 1.0,
			"rxBuffer": 64,
			"commandBuffer": 4,
			"supportM112": True,
			"echoOnM117": True,
			"brokenM29": True,
			"brokenResend": False,
			"supportF": False,
			"firmwareName": "Virtual Marlin 1.0",
			"sharedNozzle": False,
			"sendBusy": False,
			"busyInterval": 2.0,
			"simulateReset": True,
			"resetLines": ['start', 'Marlin: Virtual Marlin!', '\x80', 'SD card ok'],
			"preparedOks": [],
			"okFormatString": "ok",
			"m115FormatString": "FIRMWARE_NAME:{firmware_name} PROTOCOL_VERSION:1.0",
			"m115ReportCapabilities": True,
			"capabilities": {
				"AUTOREPORT_TEMP": True,
				"AUTOREPORT_SD_STATUS": True,
				"EMERGENCY_PARSER": True
			},
			"m114FormatString": "X:{x} Y:{y} Z:{z} E:{e[current]} Count: A:{a} B:{b} C:{c}",
			"ambientTemperature": 21.3,
			"errors": {
				"checksum_mismatch": "Checksum mismatch",
				"checksum_missing": "Missing checksum",
				"lineno_mismatch": "expected line {} got {}",
				"lineno_missing": "No Line Number with checksum, Last Line: {}",
				"maxtemp": "MAXTEMP triggered!",
				"mintemp": "MINTEMP triggered!",
				"command_unknown": "Unknown command {}"
			}
		}
	}
}
"""The default settings of the core application."""

valid_boolean_trues = CaseInsensitiveSet(True, "true", "yes", "y", "1", 1)
""" Values that are considered to be equivalent to the boolean ``True`` value, used for type conversion in various places."""


class NoSuchSettingsPath(Exception):
	pass


class InvalidSettings(Exception):
	def __init__(self, message, line=None, column=None, details=None):
		self.message = message
		self.line = line
		self.column = column
		self.details = details


class HierarchicalChainMap(ChainMap):

	def deep_dict(self, root=None):
		if root is None:
			root = self

		result = dict()
		for key, value in root.items():
			if isinstance(value, dict):
				result[key] = self.deep_dict(root=self.__class__._get_next(key, root))
			else:
				result[key] = value
		return result

	def has_path(self, path, only_local=False, only_defaults=False):
		if only_defaults:
			current = self.parents
		elif only_local:
			current = self.__class__(self.maps[0])
		else:
			current = self

		try:
			for key in path[:-1]:
				value = current[key]
				if isinstance(value, dict):
					current = self.__class__._get_next(key, current, only_local=only_local)
				else:
					return False
			return path[-1] in current
		except KeyError:
			return False

	def get_by_path(self, path, only_local=False, only_defaults=False):
		if only_defaults:
			current = self.parents
		elif only_local:
			current = self.__class__(self.maps[0])
		else:
			current = self

		for key in path[:-1]:
			value = current[key]
			if isinstance(value, dict):
				current = self.__class__._get_next(key, current, only_local=only_local)
			else:
				raise KeyError(key)

		return current[path[-1]]

	def set_by_path(self, path, value):
		current = self

		for key in path[:-1]:
			if key not in current.maps[0]:
				current.maps[0][key] = dict()
			if not isinstance(current[key], dict):
				raise KeyError(key)
			current = self.__class__._hierarchy_for_key(key, current)

		current[path[-1]] = value

	def del_by_path(self, path):
		if not path:
			raise ValueError("Invalid path")

		current = self

		for key in path[:-1]:
			if not isinstance(current[key], dict):
				raise KeyError(key)
			current = self.__class__._hierarchy_for_key(key, current)

		del current[path[-1]]


	@classmethod
	def _hierarchy_for_key(cls, key, chain):
		wrapped_mappings = list()
		for mapping in chain.maps:
			if key in mapping and mapping[key] is not None:
				wrapped_mappings.append(mapping[key])
			else:
				wrapped_mappings.append(dict())
		return HierarchicalChainMap(*wrapped_mappings)

	@classmethod
	def _get_next(cls, key, node, only_local=False):
		if isinstance(node, dict):
			return node[key]
		elif only_local and not key in node.maps[0]:
			raise KeyError(key)
		else:
			return cls._hierarchy_for_key(key, node)


class Settings(object):
	"""
	The :class:`Settings` class allows managing all of OctoPrint's settings. It takes care of initializing the settings
	directory, loading the configuration from ``config.yaml``, persisting changes to disk etc and provides access
	methods for getting and setting specific values from the overall settings structure via paths.

	A general word on the concept of paths, since they play an important role in OctoPrint's settings management. A
	path is basically a list or tuple consisting of keys to follow down into the settings (which are basically like
	a ``dict``) in order to set or retrieve a specific value (or more than one). For example, for a settings
	structure like the following::

	    serial:
	        port: "/dev/ttyACM0"
	        baudrate: 250000
	        timeouts:
	            communication: 20.0
	            temperature: 5.0
	            sdStatus: 1.0
	            connection: 10.0
	    server:
	        host: "0.0.0.0"
	        port: 5000

	the following paths could be used:

	========================================== ============================================================================
	Path                                       Value
	========================================== ============================================================================
	``["serial", "port"]``                     ::

	                                               "/dev/ttyACM0"

	``["serial", "timeout"]``                  ::

	                                               communication: 20.0
	                                               temperature: 5.0
	                                               sdStatus: 1.0
	                                               connection: 10.0

	``["serial", "timeout", "temperature"]``   ::

	                                               5.0

	``["server", "port"]``                     ::

	                                               5000

	========================================== ============================================================================

	However, these would be invalid paths: ``["key"]``, ``["serial", "port", "value"]``, ``["server", "host", 3]``.
	"""

	def __init__(self, configfile=None, basedir=None):
		self._logger = logging.getLogger(__name__)

		self._basedir = None

		assert(isinstance(default_settings, dict))

		self._map = HierarchicalChainMap(dict(), default_settings)

		self._config = None
		self._dirty = False
		self._dirty_time = 0
		self._mtime = None

		self._get_preprocessors = dict(
			controls=self._process_custom_controls
		)
		self._set_preprocessors = dict()

		self._init_basedir(basedir)

		if configfile is not None:
			self._configfile = configfile
		else:
			self._configfile = os.path.join(self._basedir, "config.yaml")
		self.load(migrate=True)

		apikey = self.get(["api", "key"])
		if not apikey or apikey == "n/a":
			self.generateApiKey()

		self._script_env = self._init_script_templating()

		self.sanity_check_folders(folders=["logs", ])

	def _init_basedir(self, basedir):
		if basedir is not None:
			self._basedir = basedir
		else:
			self._basedir = _default_basedir(_APPNAME)

		if not os.path.isdir(self._basedir):
			try:
				os.makedirs(self._basedir)
			except Exception:
				self._logger.fatal("Could not create basefolder at {}. This is a fatal error, OctoPrint "
				                   "can't run without a writable base folder.".format(self._basedir), exc_info=1)
				raise

	def sanity_check_folders(self, folders=None):
		if folders is None:
			folders = default_settings["folder"].keys()
		for folder in folders:
			self.getBaseFolder(folder, check_writable=True, deep_check_writable=True, log_error=True)

	def _get_default_folder(self, type):
		folder = default_settings["folder"][type]
		if folder is None:
			folder = os.path.join(self._basedir, type.replace("_", os.path.sep))
		return folder

	def _init_script_templating(self):
		from jinja2 import Environment, BaseLoader, ChoiceLoader, TemplateNotFound
		from jinja2.nodes import Include
		from jinja2.ext import Extension

		from octoprint.util.jinja import FilteredFileSystemLoader

		class SnippetExtension(Extension):
			tags = {"snippet"}
			fields = Include.fields

			def parse(self, parser):
				node = parser.parse_include()
				if not node.template.value.startswith("/"):
					node.template.value = "snippets/" + node.template.value
				return node

		class SettingsScriptLoader(BaseLoader):
			def __init__(self, s):
				self._settings = s

			def get_source(self, environment, template):
				parts = template.split("/")
				if not len(parts):
					raise TemplateNotFound(template)

				script = self._settings.get(["scripts"], merged=True)
				for part in parts:
					if isinstance(script, dict) and part in script:
						script = script[part]
					else:
						raise TemplateNotFound(template)
				source = script
				if source is None:
					raise TemplateNotFound(template)
				mtime = self._settings._mtime
				return source, None, lambda: mtime == self._settings.last_modified

			def list_templates(self):
				scripts = self._settings.get(["scripts"], merged=True)
				return self._get_templates(scripts)

			def _get_templates(self, scripts):
				templates = []
				for key in scripts:
					if isinstance(scripts[key], dict):
						templates += list(map(lambda x: key + "/" + x, self._get_templates(scripts[key])))
					elif isinstance(scripts[key], basestring):
						templates.append(key)
				return templates

		class SelectLoader(BaseLoader):
			def __init__(self, default, mapping, sep=":"):
				self._default = default
				self._mapping = mapping
				self._sep = sep

			def get_source(self, environment, template):
				if self._sep in template:
					prefix, name = template.split(self._sep, 1)
					if not prefix in self._mapping:
						raise TemplateNotFound(template)
					return self._mapping[prefix].get_source(environment, name)
				return self._default.get_source(environment, template)

			def list_templates(self):
				return self._default.list_templates()

		class RelEnvironment(Environment):
			def __init__(self, prefix_sep=":", *args, **kwargs):
				Environment.__init__(self, *args, **kwargs)
				self._prefix_sep = prefix_sep

			def join_path(self, template, parent):
				prefix, name = self._split_prefix(template)

				if name.startswith("/"):
					return self._join_prefix(prefix, name[1:])
				else:
					_, parent_name = self._split_prefix(parent)
					parent_base = parent_name.split("/")[:-1]
					return self._join_prefix(prefix, "/".join(parent_base) + "/" + name)

			def _split_prefix(self, template):
				if self._prefix_sep in template:
					return template.split(self._prefix_sep, 1)
				else:
					return "", template

			def _join_prefix(self, prefix, template):
				if len(prefix):
					return prefix + self._prefix_sep + template
				else:
					return template

		path_filter = lambda path: not is_hidden_path(path)
		file_system_loader = FilteredFileSystemLoader(self.getBaseFolder("scripts"),
		                                              path_filter=path_filter)
		settings_loader = SettingsScriptLoader(self)
		choice_loader = ChoiceLoader([file_system_loader, settings_loader])
		select_loader = SelectLoader(choice_loader,
		                             dict(bundled=settings_loader,
		                                  file=file_system_loader))
		return RelEnvironment(loader=select_loader, extensions=[SnippetExtension])

	def _get_script_template(self, script_type, name, source=False):
		from jinja2 import TemplateNotFound

		template_name = script_type + "/" + name
		try:
			if source:
				template_name, _, _ = self._script_env.loader.get_source(self._script_env, template_name)
				return template_name
			else:
				return self._script_env.get_template(template_name)
		except TemplateNotFound:
			return None
		except Exception:
			self._logger.exception("Exception while trying to resolve template {template_name}".format(**locals()))
			return None

	def _get_scripts(self, script_type):
		return self._script_env.list_templates(filter_func=lambda x: x.startswith(script_type+"/"))

	def _process_custom_controls(self, controls):
		def process_control(c):
			# shallow copy
			result = dict(c)

			if "regex" in result and "template" in result:
				# if it's a template matcher, we need to add a key to associate with the matcher output
				import hashlib
				key_hash = hashlib.md5()
				key_hash.update(result["regex"].encode('utf-8'))
				result["key"] = key_hash.hexdigest()

				template_key_hash = hashlib.md5()
				template_key_hash.update(result["template"].encode('utf-8'))
				result["template_key"] = template_key_hash.hexdigest()

			elif "children" in result:
				# if it has children we need to process them recursively
				result["children"] = list(map(process_control, [child for child in result["children"] if child is not None]))

			return result

		return list(map(process_control, controls))

	@property
	def effective(self):
		return self._map.deep_dict()

	@property
	def effective_yaml(self):
		import yaml
		return yaml.safe_dump(self.effective)

	@property
	def effective_hash(self):
		import hashlib
		hash = hashlib.md5()
		hash.update(self.effective_yaml.encode('utf-8'))
		return hash.hexdigest()

	@property
	def config_yaml(self):
		import yaml
		return yaml.safe_dump(self._config)

	@property
	def config_hash(self):
		import hashlib
		hash = hashlib.md5()
		hash.update(self.config_yaml.encode('utf-8'))
		return hash.hexdigest()

	@property
	def _config(self):
		return self._map.maps[0]

	@_config.setter
	def _config(self, value):
		self._map.maps[0] = value

	@property
	def _overlay_maps(self):
		if len(self._map.maps) > 2:
			return self._map.maps[1:-1]
		else:
			return []

	@property
	def _default_map(self):
		return self._map.maps[-1]

	@property
	def last_modified(self):
		"""
		Returns:
		    (int) The last modification time of the configuration file.
		"""
		stat = os.stat(self._configfile)
		return stat.st_mtime

	@property
	def last_modified_or_made_dirty(self):
		return max(self.last_modified, self._dirty_time)

	#~~ load and save

	def load(self, migrate=False):
		if os.path.exists(self._configfile) and os.path.isfile(self._configfile):
			with io.open(self._configfile, 'rt', encoding='utf-8', errors='replace') as f:
				try:
					self._config = yaml.safe_load(f)
					self._mtime = self.last_modified

				except yaml.YAMLError as e:
					details = e.message

					if hasattr(e, "problem_mark"):
						line = e.problem_mark.line
						column = e.problem_mark.column
					else:
						line = None
						column = None

					raise InvalidSettings("Invalid YAML file: {}".format(self._configfile),
					                      details=details,
					                      line=line,
					                      column=column)

		# changed from else to handle cases where the file exists, but is empty / 0 bytes
		if not self._config or not isinstance(self._config, dict):
			self._config = dict()

		if migrate:
			self._migrate_config()

	def load_overlay(self, overlay, migrate=True):
		config = None

		if callable(overlay):
			try:
				overlay = overlay(self)
			except Exception:
				self._logger.exception("Error loading overlay from callable")
				return

		if isinstance(overlay, basestring):
			if os.path.exists(overlay) and os.path.isfile(overlay):
				with io.open(overlay, 'rt', encoding='utf-8', errors='replace') as f:
					config = yaml.safe_load(f)
		elif isinstance(overlay, dict):
			config = overlay
		else:
			raise ValueError("Overlay must be either a path to a yaml file or a dictionary")

		if not isinstance(config, dict):
			raise ValueError("Configuration data must be a dict but is a {}".format(config.__class__))

		if migrate:
			self._migrate_config(config)
		return config

	def add_overlay(self, overlay, at_end=False):
		if at_end:
			pos = len(self._map.maps) - 1
			self._map.maps.insert(pos, overlay)
		else:
			self._map.maps.insert(1, overlay)

	def _migrate_config(self, config=None, persist=False):
		if config is None:
			config = self._config
			persist = True

		dirty = False

		migrators = (
			self._migrate_event_config,
			self._migrate_reverse_proxy_config,
			self._migrate_printer_parameters,
			self._migrate_gcode_scripts,
			self._migrate_core_system_commands,
			self._migrate_serial_features,
			self._migrate_resend_without_ok,
			self._migrate_string_temperature_profile_values,
			self._migrate_blocked_commands
		)

		for migrate in migrators:
			dirty = migrate(config) or dirty

		if dirty and persist:
			self.save(force=True)

	def _migrate_gcode_scripts(self, config):
		"""
		Migrates an old development version of gcode scripts to the new template based format.

		Added in 1.2.0
		"""

		dirty = False
		if "scripts" in config:
			if "gcode" in config["scripts"]:
				if "templates" in config["scripts"]["gcode"]:
					del config["scripts"]["gcode"]["templates"]

				replacements = dict(
					disable_steppers="M84",
					disable_hotends="{% snippet 'disable_hotends' %}",
					disable_bed="M140 S0",
					disable_fan="M106 S0"
				)

				for name, script in config["scripts"]["gcode"].items():
					self.saveScript("gcode", name, script.format(**replacements))
			del config["scripts"]
			dirty = True
		return dirty

	def _migrate_printer_parameters(self, config):
		"""
		Migrates the old "printer > parameters" data structure to the new printer profile mechanism.

		Added in 1.2.0
		"""
		default_profile = config["printerProfiles"]["defaultProfile"] if "printerProfiles" in config and "defaultProfile" in config["printerProfiles"] else dict()
		dirty = False

		if "printerParameters" in config:
			printer_parameters = config["printerParameters"]

			if "movementSpeed" in printer_parameters or "invertAxes" in printer_parameters:
				dirty = True
				default_profile["axes"] = dict(x=dict(), y=dict(), z=dict(), e=dict())
				if "movementSpeed" in printer_parameters:
					for axis in ("x", "y", "z", "e"):
						if axis in printer_parameters["movementSpeed"]:
							default_profile["axes"][axis]["speed"] = printer_parameters["movementSpeed"][axis]
					del config["printerParameters"]["movementSpeed"]
				if "invertedAxes" in printer_parameters:
					for axis in ("x", "y", "z", "e"):
						if axis in printer_parameters["invertedAxes"]:
							default_profile["axes"][axis]["inverted"] = True
					del config["printerParameters"]["invertedAxes"]

			if "numExtruders" in printer_parameters or "extruderOffsets" in printer_parameters:
				dirty = True
				if not "extruder" in default_profile:
					default_profile["extruder"] = dict()

				if "numExtruders" in printer_parameters:
					default_profile["extruder"]["count"] = printer_parameters["numExtruders"]
					del config["printerParameters"]["numExtruders"]
				if "extruderOffsets" in printer_parameters:
					extruder_offsets = []
					for offset in printer_parameters["extruderOffsets"]:
						if "x" in offset and "y" in offset:
							extruder_offsets.append((offset["x"], offset["y"]))
					default_profile["extruder"]["offsets"] = extruder_offsets
					del config["printerParameters"]["extruderOffsets"]

			if "bedDimensions" in printer_parameters:
				dirty = True
				bed_dimensions = printer_parameters["bedDimensions"]
				if not "volume" in default_profile:
					default_profile["volume"] = dict()

				if "circular" in bed_dimensions and "r" in bed_dimensions and bed_dimensions["circular"]:
					default_profile["volume"]["formFactor"] = "circular"
					default_profile["volume"]["width"] = 2 * bed_dimensions["r"]
					default_profile["volume"]["depth"] = default_profile["volume"]["width"]
				elif "x" in bed_dimensions or "y" in bed_dimensions:
					default_profile["volume"]["formFactor"] = "rectangular"
					if "x" in bed_dimensions:
						default_profile["volume"]["width"] = bed_dimensions["x"]
					if "y" in bed_dimensions:
						default_profile["volume"]["depth"] = bed_dimensions["y"]
				del config["printerParameters"]["bedDimensions"]

		if dirty:
			if not "printerProfiles" in config:
				config["printerProfiles"] = dict()
			config["printerProfiles"]["defaultProfile"] = default_profile
		return dirty

	def _migrate_reverse_proxy_config(self, config):
		"""
		Migrates the old "server > baseUrl" and "server > scheme" configuration entries to
		"server > reverseProxy > prefixFallback" and "server > reverseProxy > schemeFallback".

		Added in 1.2.0
		"""
		if "server" in config and ("baseUrl" in config["server"] or "scheme" in config["server"]):
			prefix = ""
			if "baseUrl" in config["server"]:
				prefix = config["server"]["baseUrl"]
				del config["server"]["baseUrl"]

			scheme = ""
			if "scheme" in config["server"]:
				scheme = config["server"]["scheme"]
				del config["server"]["scheme"]

			if not "reverseProxy" in config["server"] or not isinstance(config["server"]["reverseProxy"], dict):
				config["server"]["reverseProxy"] = dict()
			if prefix:
				config["server"]["reverseProxy"]["prefixFallback"] = prefix
			if scheme:
				config["server"]["reverseProxy"]["schemeFallback"] = scheme
			self._logger.info("Migrated reverse proxy configuration to new structure")
			return True
		else:
			return False

	def _migrate_event_config(self, config):
		"""
		Migrates the old event configuration format of type "events > gcodeCommandTrigger" and
		"event > systemCommandTrigger" to the new events format.

		Added in 1.2.0
		"""
		if "events" in config and ("gcodeCommandTrigger" in config["events"] or "systemCommandTrigger" in config["events"]):
			self._logger.info("Migrating config (event subscriptions)...")

			# migrate event hooks to new format
			placeholderRe = re.compile(r"%\((.*?)\)s")

			eventNameReplacements = {
				"ClientOpen": "ClientOpened",
				"TransferStart": "TransferStarted"
			}
			payloadDataReplacements = {
				"Upload": {"data": "{file}", "filename": "{file}"},
				"Connected": {"data": "{port} at {baudrate} baud"},
				"FileSelected": {"data": "{file}", "filename": "{file}"},
				"TransferStarted": {"data": "{remote}", "filename": "{remote}"},
				"TransferDone": {"data": "{remote}", "filename": "{remote}"},
				"ZChange": {"data": "{new}"},
				"CaptureStart": {"data": "{file}"},
				"CaptureDone": {"data": "{file}"},
				"MovieDone": {"data": "{movie}", "filename": "{gcode}"},
				"Error": {"data": "{error}"},
				"PrintStarted": {"data": "{file}", "filename": "{file}"},
				"PrintDone": {"data": "{file}", "filename": "{file}"},
			}

			def migrateEventHook(event, command):
				# migrate placeholders
				command = placeholderRe.sub("{__\\1}", command)

				# migrate event names
				if event in eventNameReplacements:
					event = eventNameReplacements["event"]

				# migrate payloads to more specific placeholders
				if event in payloadDataReplacements:
					for key in payloadDataReplacements[event]:
						command = command.replace("{__%s}" % key, payloadDataReplacements[event][key])

				# return processed tuple
				return event, command

			disableSystemCommands = False
			if "systemCommandTrigger" in config["events"] and "enabled" in config["events"]["systemCommandTrigger"]:
				disableSystemCommands = not config["events"]["systemCommandTrigger"]["enabled"]

			disableGcodeCommands = False
			if "gcodeCommandTrigger" in config["events"] and "enabled" in config["events"]["gcodeCommandTrigger"]:
				disableGcodeCommands = not config["events"]["gcodeCommandTrigger"]["enabled"]

			disableAllCommands = disableSystemCommands and disableGcodeCommands
			newEvents = {
				"enabled": not disableAllCommands,
				"subscriptions": []
			}

			if "systemCommandTrigger" in config["events"] and "subscriptions" in config["events"]["systemCommandTrigger"]:
				for trigger in config["events"]["systemCommandTrigger"]["subscriptions"]:
					if not ("event" in trigger and "command" in trigger):
						continue

					newTrigger = {"type": "system"}
					if disableSystemCommands and not disableAllCommands:
						newTrigger["enabled"] = False

					newTrigger["event"], newTrigger["command"] = migrateEventHook(trigger["event"], trigger["command"])
					newEvents["subscriptions"].append(newTrigger)

			if "gcodeCommandTrigger" in config["events"] and "subscriptions" in config["events"]["gcodeCommandTrigger"]:
				for trigger in config["events"]["gcodeCommandTrigger"]["subscriptions"]:
					if not ("event" in trigger and "command" in trigger):
						continue

					newTrigger = {"type": "gcode"}
					if disableGcodeCommands and not disableAllCommands:
						newTrigger["enabled"] = False

					newTrigger["event"], newTrigger["command"] = migrateEventHook(trigger["event"], trigger["command"])
					newTrigger["command"] = newTrigger["command"].split(",")
					newEvents["subscriptions"].append(newTrigger)

			config["events"] = newEvents
			self._logger.info("Migrated %d event subscriptions to new format and structure" % len(newEvents["subscriptions"]))
			return True
		else:
			return False

	def _migrate_core_system_commands(self, config):
		"""
		Migrates system commands for restart, reboot and shutdown as defined on OctoPi or
		according to the official setup guide to new core system commands to remove
		duplication.

		If server commands for action is not yet set, migrates command. Otherwise only
		deletes definition from custom system commands.

		Added in 1.3.0
		"""
		changed = False

		migration_map = dict(shutdown="systemShutdownCommand",
		                     reboot="systemRestartCommand",
		                     restart="serverRestartCommand")

		if "system" in config and "actions" in config["system"] and isinstance(config["system"]["actions"], (list, tuple)):
			actions = config["system"]["actions"]
			to_delete = []
			for index, spec in enumerate(actions):
				action = spec.get("action")
				command = spec.get("command")
				if action is None or command is None:
					continue

				migrate_to = migration_map.get(action)
				if migrate_to is not None:
					if not "server" in config or not "commands" in config["server"] or not migrate_to in config["server"]["commands"]:
						if not "server" in config:
							config["server"] = dict()
						if not "commands" in config["server"]:
							config["server"]["commands"] = dict()
						config["server"]["commands"][migrate_to] = command
						self._logger.info("Migrated {} action to server.commands.{}".format(action, migrate_to))

					to_delete.append(index)
					self._logger.info("Deleting {} action from configured system commands, superseded by server.commands.{}".format(action, migrate_to))

			for index in reversed(to_delete):
				actions.pop(index)
				changed = True

		if changed:
			# let's make a backup of our current config, in case someone wants to roll back to an
			# earlier version and needs to recover the former system commands for that
			backup_path = self.backup("system_command_migration")
			self._logger.info("Made a copy of the current config at {} to allow recovery of manual system command configuration".format(backup_path))

		return changed

	def _migrate_serial_features(self, config):
		"""
		Migrates feature flags identified as serial specific from the feature to the serial config tree and vice versa.

		If a flag already exists in the target tree, only deletes the copy in the source tree.

		Added in 1.3.7
		"""
		changed = False

		FEATURE_TO_SERIAL = ("waitForStartOnConnect", "alwaysSendChecksum", "neverSendChecksum",
		                     "sendChecksumWithUnknownCommands", "unknownCommandsNeedAck", "sdRelativePath",
		                     "sdAlwaysAvailable", "swallowOkAfterResend", "repetierTargetTemp",
		                     "externalHeatupDetection", "supportWait", "ignoreIdenticalResends",
		                     "identicalResendsCountdown", "supportFAsCommand", "firmwareDetection",
		                     "blockWhileDwelling")
		SERIAL_TO_FEATURE = ("autoUppercaseBlacklist",)

		def migrate_key(key, source, target):
			if source in config and key in config[source]:
				if config.get(target) is None:
					# make sure we have a serial tree
					config[target] = dict()
				if key not in config[target]:
					# only copy over if it's not there yet
					config[target][key] = config[source][key]
				# delete feature flag
				del config[source][key]
				return True
			return False

		for key in FEATURE_TO_SERIAL:
			changed = migrate_key(key, "feature", "serial") or changed

		for key in SERIAL_TO_FEATURE:
			changed = migrate_key(key, "serial", "feature") or changed

		if changed:
			# let's make a backup of our current config, in case someone wants to roll back to an
			# earlier version and needs a backup of their flags
			backup_path = self.backup("serial_feature_migration")
			self._logger.info("Made a copy of the current config at {} to allow recovery of serial feature flags".format(backup_path))

		return changed

	def _migrate_resend_without_ok(self, config):
		"""
		Migrates supportResendsWithoutOk flag from boolean to ("always", "detect", "never") value range.

		True gets migrated to "always", False to "detect" (which is the new default).

		Added in 1.3.7
		"""
		if "serial" in config and "supportResendsWithoutOk" in config["serial"] \
				and config["serial"]["supportResendsWithoutOk"] not in ("always", "detect", "never"):
			value = config["serial"]["supportResendsWithoutOk"]
			if value:
				config["serial"]["supportResendsWithoutOk"] = "always"
			else:
				config["serial"]["supportResendsWithoutOk"] = "detect"
			return True
		return False

	def _migrate_string_temperature_profile_values(self, config):
		"""
		Migrates/fixes temperature profile wrongly saved with strings instead of ints as temperature values.

		Added in 1.3.8
		"""
		if "temperature" in config and "profiles" in config["temperature"]:
			profiles = config["temperature"]["profiles"]
			if any(map(lambda x: not isinstance(x.get("extruder", 0), int) or not isinstance(x.get("bed", 0), int),
			           profiles)):
				result = []
				for profile in profiles:
					try:
						profile["extruder"] = int(profile["extruder"])
						profile["bed"] = int(profile["bed"])
					except ValueError:
						pass
					result.append(profile)
				config["temperature"]["profiles"] = result
				return True
		return False

	def _migrate_blocked_commands(self, config):
		if "serial" in config and "blockM0M1" in config["serial"]:
			blockM0M1 = config["serial"]["blockM0M1"]
			blockedCommands = config["serial"].get("blockedCommands", [])
			if blockM0M1:
				blockedCommands = set(blockedCommands)
				blockedCommands.add("M0")
				blockedCommands.add("M1")
				config["serial"]["blockedCommands"] = sorted(blockedCommands)
			else:
				config["serial"]["blockedCommands"] = sorted([v for v in blockedCommands if v not in ("M0", "M1")])
			del config["serial"]["blockM0M1"]
			return True
		return False

	def backup(self, suffix=None, path=None, ext=None, hidden=False):
		import shutil

		if path is None:
			path = os.path.dirname(self._configfile)

		basename = os.path.basename(self._configfile)
		name, default_ext = os.path.splitext(basename)

		if ext is None:
			ext = default_ext

		if suffix is None and ext == default_ext:
			raise ValueError("Need a suffix or a different extension")

		if suffix is None:
			suffix = ""

		backup = os.path.join(path, "{}{}.{}{}".format("." if hidden else "", name, suffix, ext))
		shutil.copy(self._configfile, backup)
		return backup

	def save(self, force=False, trigger_event=False):
		if not self._dirty and not force:
			return False

		try:
			with atomic_write(self._configfile, mode='wt', prefix="octoprint-config-", suffix=".yaml", permissions=0o600, max_permissions=0o666) as configFile:
				yaml.safe_dump(self._config, configFile, default_flow_style=False, indent=4, allow_unicode=True)
				self._dirty = False
		except Exception:
			self._logger.exception("Error while saving config.yaml!")
			raise
		else:
			from octoprint.events import eventManager, Events

			self.load()

			if trigger_event:
				payload = dict(config_hash=self.config_hash,
				               effective_hash=self.effective_hash)
				eventManager().fire(Events.SETTINGS_UPDATED, payload=payload)

			return True

	##~~ Internal getter

	def _get_by_path(self, path, config):
		current = config
		for key in path:
			if key not in current:
				raise NoSuchSettingsPath()
			current = current[key]
		return current

	def _get_value(self, path, asdict=False, config=None, defaults=None, preprocessors=None, merged=False, incl_defaults=True, do_copy=True):
		if not path:
			raise NoSuchSettingsPath()

		if config is not None or defaults is not None:
			if config is None:
				config = self._config

			if defaults is None:
				defaults = dict(self._map.parents)

			# mappings: provided config + any intermediary parents + provided defaults + regular defaults
			mappings = [config] + self._overlay_maps + [defaults, self._default_map]
			chain = HierarchicalChainMap(*mappings)
		else:
			chain = self._map

		if preprocessors is None:
			preprocessors = self._get_preprocessors

		preprocessor = None
		try:
			preprocessor = self._get_by_path(path, preprocessors)
		except NoSuchSettingsPath:
			pass

		parent_path = path[:-1]
		last = path[-1]

		if not isinstance(last, (list, tuple)):
			keys = [last]
		else:
			keys = last

		if asdict:
			results = dict()
		else:
			results = list()

		for key in keys:
			try:
				value = chain.get_by_path(parent_path + [key], only_local=not incl_defaults)
			except KeyError:
				raise NoSuchSettingsPath()

			if isinstance(value, dict) and merged:
				try:
					default_value = chain.get_by_path(parent_path + [key], only_defaults=True)
					if default_value is not None:
						value = dict_merge(default_value, value)
				except KeyError:
					raise NoSuchSettingsPath()

			if preprocessors is not None:
				try:
					preprocessor = self._get_by_path(path, preprocessors)
				except Exception:
					pass

				if callable(preprocessor):
					value = preprocessor(value)

			if do_copy:
				if isinstance(value, KeysView):
					value = list(value)
				value = copy.deepcopy(value)

			if asdict:
				results[key] = value
			else:
				results.append(value)

		if not isinstance(last, (list, tuple)):
			if asdict:
				return list(results.values()).pop()
			else:
				return results.pop()
		else:
			return results

	#~~ has

	def has(self, path, **kwargs):
		try:
			self._get_value(path, **kwargs)
		except NoSuchSettingsPath:
			return False
		else:
			return True

	#~~ getter

	def get(self, path, **kwargs):
		error_on_path = kwargs.pop("error_on_path", False)
		validator = kwargs.pop("validator", None)
		fallback = kwargs.pop("fallback", None)

		def process():
			try:
				return self._get_value(path, **kwargs)
			except NoSuchSettingsPath:
				if error_on_path:
					raise
				return None

		result = process()
		if callable(validator) and not validator(result):
			result = fallback
		return result

	def getInt(self, path, **kwargs):
		minimum = kwargs.pop("min", None)
		maximum = kwargs.pop("max", None)

		value = self.get(path, **kwargs)
		if value is None:
			return None

		try:
			intValue = int(value)

			if minimum is not None and intValue < minimum:
				return minimum
			elif maximum is not None and intValue > maximum:
				return maximum
			else:
				return intValue
		except ValueError:
			self._logger.warning("Could not convert %r to a valid integer when getting option %r" % (value, path))
			return None

	def getFloat(self, path, **kwargs):
		minimum = kwargs.pop("min", None)
		maximum = kwargs.pop("max", None)

		value = self.get(path, **kwargs)
		if value is None:
			return None

		try:
			floatValue = float(value)

			if minimum is not None and floatValue < minimum:
				return minimum
			elif maximum is not None and floatValue > maximum:
				return maximum
			else:
				return floatValue
		except ValueError:
			self._logger.warning("Could not convert %r to a valid integer when getting option %r" % (value, path))
			return None

	def getBoolean(self, path, **kwargs):
		value = self.get(path, **kwargs)
		if value is None:
			return None
		if isinstance(value, bool):
			return value
		if isinstance(value, (int, float)):
			return value != 0
		if isinstance(value, basestring):
			return value.lower() in valid_boolean_trues
		return value is not None

	def getBaseFolder(self, type, create=True, allow_fallback=True, log_error=False, check_writable=True, deep_check_writable=False):
		if type != "base" and type not in default_settings["folder"]:
			return None

		if type == "base":
			return self._basedir

		folder = self.get(["folder", type])
		default_folder = self._get_default_folder(type)
		if folder is None:
			folder = default_folder

		try:
			_validate_folder(folder, create=create, check_writable=check_writable, deep_check_writable=deep_check_writable, log_error=log_error)
		except Exception:
			if folder != default_folder and allow_fallback:
				if log_error:
					self._logger.error("Invalid configured {} folder at {}, attempting to "
					                   "fall back on default folder at {}".format(type,
					                                                              folder,
					                                                              default_folder))
				_validate_folder(default_folder, create=create, check_writable=check_writable, deep_check_writable=deep_check_writable, log_error=log_error)
				folder = default_folder

				try:
					del self._config["folder"][type]
					if not len(self._config["folder"]):
						del self._config["folder"]
					self._dirty = True
					self._dirty_time = time.time()
					self.save()
				except KeyError:
					pass
			else:
				raise

		return folder

	def listScripts(self, script_type):
		return list(map(lambda x: x[len(script_type + "/"):], filter(lambda x: x.startswith(script_type + "/"), self._get_scripts(script_type))))

	def loadScript(self, script_type, name, context=None, source=False):
		if context is None:
			context = dict()
		context.update(dict(script=dict(type=script_type, name=name)))

		template = self._get_script_template(script_type, name, source=source)
		if template is None:
			return None

		if source:
			script = template
		else:
			try:
				script = template.render(**context)
			except Exception:
				self._logger.exception("Exception while trying to render script {script_type}:{name}".format(**locals()))
				return None

		return script

	#~~ remove

	def remove(self, path, config=None, error_on_path=False):
		if not path:
			if error_on_path:
				raise NoSuchSettingsPath()
			return

		if config is not None:
			mappings = [config] + self._overlay_maps + [self._default_map]
			chain = HierarchicalChainMap(*mappings)
		else:
			chain = self._map

		try:
			chain.del_by_path(path)
			self._dirty = True
			self._dirty_time = time.time()
		except KeyError:
			if error_on_path:
				raise NoSuchSettingsPath()
			pass

	#~~ setter

	def set(self, path, value, force=False, defaults=None, config=None, preprocessors=None, error_on_path=False, *args, **kwargs):
		if not path:
			if error_on_path:
				raise NoSuchSettingsPath()
			return

		if self._mtime is not None and self.last_modified != self._mtime:
			self.load()

		if config is not None or defaults is not None:
			if config is None:
				config = self._config

			if defaults is None:
				defaults = dict(self._map.parents)

			chain = HierarchicalChainMap(config, defaults)
		else:
			chain = self._map

		if preprocessors is None:
			preprocessors = self._set_preprocessors

		preprocessor = None
		try:
			preprocessor = self._get_by_path(path, preprocessors)
		except NoSuchSettingsPath:
			pass

		if callable(preprocessor):
			value = preprocessor(value)

		try:
			current = chain.get_by_path(path)
		except KeyError:
			current = None

		try:
			default_value = chain.get_by_path(path, only_defaults=True)
		except KeyError:
			if error_on_path:
				raise NoSuchSettingsPath()
			default_value = None

		in_local = chain.has_path(path, only_local=True)
		in_defaults = chain.has_path(path, only_defaults=True)

		if not force and in_defaults and in_local and default_value == value:
			try:
				chain.del_by_path(path)
				self._dirty = True
				self._dirty_time = time.time()
			except KeyError:
				if error_on_path:
					raise NoSuchSettingsPath()
				pass
		elif force or (not in_local and in_defaults and default_value != value) or (in_local and current != value):
			if value is None and in_local:
				chain.del_by_path(path)
			else:
				chain.set_by_path(path, value)
			self._dirty = True
			self._dirty_time = time.time()

	def setInt(self, path, value, **kwargs):
		if value is None:
			self.set(path, None, **kwargs)
			return

		minimum = kwargs.pop("min", None)
		maximum = kwargs.pop("max", None)

		try:
			intValue = int(value)

			if minimum is not None and intValue < minimum:
				intValue = minimum
			if maximum is not None and intValue > maximum:
				intValue = maximum
		except ValueError:
			self._logger.warning("Could not convert %r to a valid integer when setting option %r" % (value, path))
			return

		self.set(path, intValue, **kwargs)

	def setFloat(self, path, value, **kwargs):
		if value is None:
			self.set(path, None, **kwargs)
			return

		minimum = kwargs.pop("min", None)
		maximum = kwargs.pop("max", None)

		try:
			floatValue = float(value)

			if minimum is not None and floatValue < minimum:
				floatValue = minimum
			if maximum is not None and floatValue > maximum:
				floatValue = maximum
		except ValueError:
			self._logger.warning("Could not convert %r to a valid integer when setting option %r" % (value, path))
			return

		self.set(path, floatValue, **kwargs)

	def setBoolean(self, path, value, **kwargs):
		if value is None or isinstance(value, bool):
			self.set(path, value, **kwargs)
		elif isinstance(value, basestring) and value.lower() in valid_boolean_trues:
			self.set(path, True, **kwargs)
		else:
			self.set(path, False, **kwargs)

	def setBaseFolder(self, type, path, force=False, validate=True):
		if type not in default_settings["folder"]:
			return None

		currentPath = self.getBaseFolder(type)
		defaultPath = self._get_default_folder(type)
		if (path is None or path == defaultPath) and "folder" in self._config and type in self._config["folder"]:
			del self._config["folder"][type]
			if not self._config["folder"]:
				del self._config["folder"]
			self._dirty = True
			self._dirty_time = time.time()
		elif (path != currentPath and path != defaultPath) or force:
			if validate:
				_validate_folder(path, check_writable=True, deep_check_writable=True)

			if "folder" not in self._config:
				self._config["folder"] = {}
			self._config["folder"][type] = path
			self._dirty = True
			self._dirty_time = time.time()

	def saveScript(self, script_type, name, script):
		script_folder = self.getBaseFolder("scripts")
		filename = os.path.realpath(os.path.join(script_folder, script_type, name))
		if not filename.startswith(os.path.realpath(script_folder)):
			# oops, jail break, that shouldn't happen
			raise ValueError("Invalid script path to save to: {filename} (from {script_type}:{name})".format(**locals()))

		path, _ = os.path.split(filename)
		if not os.path.exists(path):
			os.makedirs(path)
		with atomic_write(filename, mode='wt', max_permissions=0o666) as f:
			f.write(script)

	def generateApiKey(self):
		apikey = generate_api_key()
		self.set(["api", "key"], apikey)
		self.save(force=True)
		return apikey

	def deleteApiKey(self):
		self.set(["api", "key"], None)
		self.save(force=True)


def _default_basedir(applicationName):
	# taken from http://stackoverflow.com/questions/1084697/how-do-i-store-desktop-application-data-in-a-cross-platform-way-for-python
	if sys.platform == "darwin":
		import appdirs
		return appdirs.user_data_dir(applicationName, "")
	elif sys.platform == "win32":
		return os.path.join(os.environ["APPDATA"], applicationName)
	else:
		return os.path.expanduser(os.path.join("~", "." + applicationName.lower()))


def _validate_folder(folder, create=True, check_writable=True, deep_check_writable=False, log_error=False):
	logger = logging.getLogger(__name__)

	if not os.path.exists(folder):
		if os.path.islink(folder):
			# broken symlink, see #2644
			raise IOError("Folder at {} appears to be a broken symlink".format(folder))

		elif create:
			# non existing, but we are allowed to create it
			try:
				os.makedirs(folder)
			except Exception:
				if log_error:
					logger.exception("Could not create {}".format(folder))
				raise IOError("Folder for type {} at {} does not exist and creation failed".format(type, folder))

		else:
			# not extisting, not allowed to create it
			raise IOError("No such folder: {}".format(folder))

	elif os.path.isfile(folder):
		# hardening against misconfiguration, see #1953
		raise IOError("Expected a folder at {} but found a file instead".format(folder))

	elif check_writable:
		# make sure we can also write into the folder
		error = "Folder at {} doesn't appear to be writable, please fix its permissions".format(folder)
		if not os.access(folder, os.W_OK):
			raise IOError(error)

		elif deep_check_writable:
			# try to write a file to the folder - on network shares that might be the only reliable way
			# to determine whether things are *actually* writable
			testfile = os.path.join(folder, ".testballoon.txt")
			try:
				with io.open(testfile, 'wt', encoding='utf-8') as f:
					f.write("test")
				os.remove(testfile)
			except Exception:
				if log_error:
					logger.exception("Could not write test file to {}".format(folder))
				raise IOError(error)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2019 The OctoPrint Project - Released under terms of the AGPLv3 License"

from flask_babel import gettext
from octoprint.util.version import get_comparable_version
from . import Check, AuthorCheck

class FirmwareUnsafeChecks(object):
	@classmethod
	def as_dict(cls):
		return dict(checks=(AnetA8Check(), AnycubicCheck(), CrealityCR10sCheck(), CrealityEnder3Check(),
	                        MalyanM200Check(), Micro3DIMECheck(), Micro3DStockCheck(), RepetierBefore092Check(),
	                        ThermalProtectionCapCheck()),
		            message=gettext("Your printer's firmware is known to lack mandatory safety features (e.g. "
		                            "thermal runaway protection). This is a fire risk."))

class AnetA8Check(Check):
	"""
	Anet A8 stock firmware

	Identified through firmware name "ANET_A8_".
	"""
	name = "anet_a8"

	def m115(self, name, data):
		self._triggered = name and name.lower().startswith("anet_a8_")
		self._active = False

class AnycubicCheck(Check):
	"""
	Anycubic MEGA stock firmware

	Identified through "Author: (Jolly, xxxxxxxx.CO.)" or "| Author: (**Jolly, xxxxxxxx.CO.**)" in startup messages
	combined with "echo:Vx.y.z" in startup messages, with x.y.z < 1.1.2.
	"""
	name = "anycubic"

	AUTHOR = "| Author: ".lower()
	VERSION = "echo:V"

	CRITICAL_AUTHOR1 = "| Author: (Jolly, xxxxxxxx.CO.)".lower()
	CRITICAL_AUTHOR2 = "| Author: (**Jolly, xxxxxxxx.CO.**)".lower()

	FIXED_VERSION = get_comparable_version("1.1.2")

	def __init__(self):
		Check.__init__(self)
		self._author_matches = None
		self._version_matches = None

	def received(self, line):
		if not line:
			return

		lower_line = line.lower()
		if self.AUTHOR in lower_line:
			self._author_matches = self.CRITICAL_AUTHOR1 in lower_line or self.CRITICAL_AUTHOR2 in lower_line
		elif line.startswith(self.VERSION):
			self._version_matches = self._broken_version(line)
		else:
			return

		self._evaluate()

	def _broken_version(self, line):
		version_str = line[len(self.VERSION):]
		version = get_comparable_version(version_str, base=True)
		if version is not None and version < self.FIXED_VERSION:
			return True
		else:
			return False

	def _evaluate(self):
		if self._author_matches is None or self._version_matches is None:
			return
		self._triggered = self._author_matches and self._version_matches
		self._active = False

	def reset(self):
		Check.reset(self)
		self._author_matches = None
		self._version_matches = None

class CrealityCR10sCheck(AuthorCheck):
	"""
	Creality CR10s

	Identified through " | Author: (CR-10Slanguage)" in startup messages.
	"""
	name = "creality_cr10s"
	authors = (" | Author: (CR-10Slanguage)".lower(),)

class CrealityEnder3Check(AuthorCheck):
	"""
	Creality Ender3

	Identified through " | Author: (Ender3)" in startup messages.
	"""
	name = "creality_ender3"
	authors = (" | Author: (Ender3)".lower(),)

class MalyanM200Check(Check):
	"""
	Malyan M200 stock firmware prior to version 4.0

	Identified through firmware name "Malyan*", model "M200" and version < 4.0.
	"""
	name = "malyan_m200"

	FIXED_VERSION = get_comparable_version("4.0")

	def m115(self, name, data):
		self._triggered = name and name.lower().startswith("malyan") and data.get("MODEL") == "M200" and get_comparable_version(data.get("VER", "0")) < self.FIXED_VERSION
		self._active = False

class Micro3DIMECheck(Check):
	"""
	Micro3D with IME firmware

	Identified through firmware name "iME*".
	"""
	name = "micro3d_ime"

	def m115(self, name, data):
		self._triggered = name and name.lower().startswith("ime")
		self._active = False

class Micro3DStockCheck(Check):
	"""
	Micro3D with IME firmware

	Identified through firmware name "Micro3D*".
	"""
	name = "micro3d"

	def m115(self, name, data):
		self._triggered = name and name.lower().startswith("micro3d")
		self._active = False

class RepetierBefore092Check(Check):
	"""
	Repetier firmware prior to version 0.92

	Identified through firmware name "Repetier_x.y.z" with x.y.z < 0.92
	"""
	name = "repetier_before_092"

	FIXED_VERSION = get_comparable_version("0.92")

	def m115(self, name, data):
		if name and name.lower().startswith("repetier"):
			version = self._extract_repetier_version(name)
			self._triggered = version is not None and version < self.FIXED_VERSION
		self._active = False

	def _extract_repetier_version(self, name):
		"""
		Extracts the Repetier version number from the firmware name.

		Example: "Repetier_0.91" => 0.91
		"""
		version = None
		if "_" in name:
			_, version = name.split("_", 1)
			version = get_comparable_version(version, base=True)
		return version

class ThermalProtectionCapCheck(Check):
	"""
	Firmware reporting disabled THERMAL_PROTECTION capability
	"""
	name = "capability"

	def cap(self, cap, enabled):
		if cap == "THERMAL_PROTECTION":
			self._triggered = not enabled
			self._active = False



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>, Lars Norpchen"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

import datetime
import logging
import subprocess
try:
	import queue
except ImportError:
	import Queue as queue
import threading
import collections
import re

from octoprint.settings import settings
import octoprint.plugin

# singleton
_instance = None


def all_events():
	return [getattr(Events, name) for name in Events.__dict__ if not name.startswith("_") and not name in ("register_event",)]


class Events(object):
	# server
	STARTUP = "Startup"
	SHUTDOWN = "Shutdown"
	CONNECTIVITY_CHANGED = "ConnectivityChanged"

	# connect/disconnect to printer
	CONNECTING = "Connecting"
	CONNECTED = "Connected"
	DISCONNECTING = "Disconnecting"
	DISCONNECTED = "Disconnected"

	# State changes
	PRINTER_STATE_CHANGED = "PrinterStateChanged"
	PRINTER_RESET = "PrinterReset"

	# connect/disconnect by client
	CLIENT_OPENED = "ClientOpened"
	CLIENT_CLOSED = "ClientClosed"
	CLIENT_AUTHED = "ClientAuthed"
	CLIENT_DEAUTHED = "ClientDeauthed"

	# user login/logout
	USER_LOGGED_IN = "UserLoggedIn"
	USER_LOGGED_OUT = "UserLoggedOut"

	# File management
	UPLOAD = "Upload"
	FILE_SELECTED = "FileSelected"
	FILE_DESELECTED = "FileDeselected"
	UPDATED_FILES = "UpdatedFiles"
	METADATA_ANALYSIS_STARTED = "MetadataAnalysisStarted"
	METADATA_ANALYSIS_FINISHED = "MetadataAnalysisFinished"
	METADATA_STATISTICS_UPDATED = "MetadataStatisticsUpdated"

	FILE_ADDED = "FileAdded"
	FILE_REMOVED = "FileRemoved"
	FOLDER_ADDED = "FolderAdded"
	FOLDER_REMOVED = "FolderRemoved"

	# SD Upload
	TRANSFER_STARTED = "TransferStarted"
	TRANSFER_DONE = "TransferDone"
	TRANSFER_FAILED = "TransferFailed"

	# print job
	PRINT_STARTED = "PrintStarted"
	PRINT_DONE = "PrintDone"
	PRINT_FAILED = "PrintFailed"
	PRINT_CANCELLING = "PrintCancelling"
	PRINT_CANCELLED = "PrintCancelled"
	PRINT_PAUSED = "PrintPaused"
	PRINT_RESUMED = "PrintResumed"
	ERROR = "Error"

	# print/gcode events
	POWER_ON = "PowerOn"
	POWER_OFF = "PowerOff"
	HOME = "Home"
	Z_CHANGE = "ZChange"
	WAITING = "Waiting"
	DWELL = "Dwelling"
	COOLING = "Cooling"
	ALERT = "Alert"
	CONVEYOR = "Conveyor"
	EJECT = "Eject"
	E_STOP = "EStop"
	POSITION_UPDATE = "PositionUpdate"
	FIRMWARE_DATA = "FirmwareData"
	TOOL_CHANGE = "ToolChange"
	REGISTERED_MESSAGE_RECEIVED = "RegisteredMessageReceived"

	# Timelapse
	CAPTURE_START = "CaptureStart"
	CAPTURE_DONE = "CaptureDone"
	CAPTURE_FAILED = "CaptureFailed"
	POSTROLL_START = "PostRollStart"
	POSTROLL_END = "PostRollEnd"
	MOVIE_RENDERING = "MovieRendering"
	MOVIE_DONE = "MovieDone"
	MOVIE_FAILED = "MovieFailed"

	# Slicing
	SLICING_STARTED = "SlicingStarted"
	SLICING_DONE = "SlicingDone"
	SLICING_FAILED = "SlicingFailed"
	SLICING_CANCELLED = "SlicingCancelled"
	SLICING_PROFILE_ADDED = "SlicingProfileAdded"
	SLICING_PROFILE_MODIFIED = "SlicingProfileModified"
	SLICING_PROFILE_DELETED = "SlicingProfileDeleted"

	# Printer Profiles
	PRINTER_PROFILE_ADDED = "PrinterProfileAdded"
	PRINTER_PROFILE_MODIFIED = "PrinterProfileModified"
	PRINTER_PROFILE_DELETED = "PrinterProfileDeleted"

	# Settings
	SETTINGS_UPDATED = "SettingsUpdated"

	@classmethod
	def register_event(cls, event, prefix=None):
		name = cls._to_identifier(event)
		if prefix:
			event = prefix + event
			name = cls._to_identifier(prefix) + name
		setattr(cls, name, event)
		return name, event

	# based on https://stackoverflow.com/a/1176023
	_first_cap_re = re.compile('(.)([A-Z][a-z]+)')
	_all_cap_re = re.compile('([a-z0-9])([A-Z])')

	@classmethod
	def _to_identifier(cls, name):
		s1 = cls._first_cap_re.sub(r'\1_\2', name)
		return cls._all_cap_re.sub(r'\1_\2', s1).upper()

def eventManager():
	global _instance
	if _instance is None:
		_instance = EventManager()
	return _instance


class EventManager(object):
	"""
	Handles receiving events and dispatching them to subscribers
	"""

	def __init__(self):
		self._registeredListeners = collections.defaultdict(list)
		self._logger = logging.getLogger(__name__)
		self._logger_fire = logging.getLogger("{}.fire".format(__name__))

		self._startup_signaled = False
		self._shutdown_signaled = False

		self._queue = queue.Queue()
		self._held_back = queue.Queue()

		self._worker = threading.Thread(target=self._work)
		self._worker.daemon = True
		self._worker.start()

	def _work(self):
		try:
			while not self._shutdown_signaled:
				event, payload = self._queue.get(True)
				if event == Events.SHUTDOWN:
					# we've got the shutdown event here, stop event loop processing after this has been processed
					self._logger.info("Processing shutdown event, this will be our last event")
					self._shutdown_signaled = True

				eventListeners = self._registeredListeners[event]
				self._logger_fire.debug("Firing event: {} (Payload: {!r})".format(event, payload))

				for listener in eventListeners:
					self._logger.debug("Sending action to {!r}".format(listener))
					try:
						listener(event, payload)
					except Exception:
						self._logger.exception("Got an exception while sending event {} (Payload: {!r}) to {}".format(event, payload, listener))

				octoprint.plugin.call_plugin(octoprint.plugin.types.EventHandlerPlugin,
				                             "on_event",
				                             args=(event, payload))
			self._logger.info("Event loop shut down")
		except Exception:
			self._logger.exception("Ooops, the event bus worker loop crashed")

	def fire(self, event, payload=None):
		"""
		Fire an event to anyone subscribed to it

		Any object can generate an event and any object can subscribe to the event's name as a string (arbitrary, but
		case sensitive) and any extra payload data that may pertain to the event.

		Callbacks must implement the signature "callback(event, payload)", with "event" being the event's name and
		payload being a payload object specific to the event.
		"""

		send_held_back = False
		if event == Events.STARTUP:
			self._logger.info("Processing startup event, this is our first event")
			self._startup_signaled = True
			send_held_back = True

		self._enqueue(event, payload)

		if send_held_back:
			self._logger.info("Adding {} events to queue that "
			                  "were held back before startup event".format(self._held_back.qsize()))
			while True:
				try:
					self._queue.put(self._held_back.get(block=False))
				except queue.Empty:
					break

	def _enqueue(self, event, payload):
		if self._startup_signaled:
			q = self._queue
		else:
			q = self._held_back

		q.put((event, payload))

	def subscribe(self, event, callback):
		"""
		Subscribe a listener to an event -- pass in the event name (as a string) and the callback object
		"""

		if callback in self._registeredListeners[event]:
			# callback is already subscribed to the event
			return

		self._registeredListeners[event].append(callback)
		self._logger.debug("Subscribed listener {!r} for event {}".format(callback, event))

	def unsubscribe (self, event, callback):
		"""
		Unsubscribe a listener from an event -- pass in the event name (as string) and the callback object
		"""

		try:
			self._registeredListeners[event].remove(callback)
		except ValueError:
			# not registered
			pass

	def join(self, timeout=None):
		self._worker.join(timeout)
		return self._worker.is_alive()


class GenericEventListener(object):
	"""
	The GenericEventListener can be subclassed to easily create custom event listeners.
	"""

	def __init__(self):
		self._logger = logging.getLogger(__name__)

	def subscribe(self, events):
		"""
		Subscribes the eventCallback method for all events in the given list.
		"""

		for event in events:
			eventManager().subscribe(event, self.eventCallback)

	def unsubscribe(self, events):
		"""
		Unsubscribes the eventCallback method for all events in the given list
		"""

		for event in events:
			eventManager().unsubscribe(event, self.eventCallback)

	def eventCallback(self, event, payload):
		"""
		Actual event callback called with name of event and optional payload. Not implemented here, override in
		child classes.
		"""
		pass


class DebugEventListener(GenericEventListener):
	def __init__(self):
		GenericEventListener.__init__(self)

		events = list(filter(lambda x: not x.startswith("__"), dir(Events)))
		self.subscribe(events)

	def eventCallback(self, event, payload):
		GenericEventListener.eventCallback(self, event, payload)
		self._logger.debug("Received event: %s (Payload: %r)" % (event, payload))


class CommandTrigger(GenericEventListener):
	def __init__(self, printer):
		GenericEventListener.__init__(self)
		self._printer = printer
		self._subscriptions = {}

		self._initSubscriptions()

	def _initSubscriptions(self):
		"""
		Subscribes all events as defined in "events > $triggerType > subscriptions" in the settings with their
		respective commands.
		"""
		if not settings().get(["events"]):
			return

		if not settings().getBoolean(["events", "enabled"]):
			return

		eventsToSubscribe = []
		subscriptions = settings().get(["events", "subscriptions"])
		for subscription in subscriptions:
			if not isinstance(subscription, dict):
				self._logger.info("Invalid subscription definition, not a dictionary: {!r}".format(subscription))
				continue

			if not "event" in subscription or not "command" in subscription \
					or not "type" in subscription or not subscription["type"] in ["system", "gcode"]:
				self._logger.info("Invalid command trigger, missing either event, type or command or type is invalid: {!r}".format(subscription))
				continue

			if "enabled" in subscription and not subscription["enabled"]:
				self._logger.info("Disabled command trigger: {!r}".format(subscription))
				continue

			events = subscription["event"]
			command = subscription["command"]
			commandType = subscription["type"]
			debug = subscription["debug"] if "debug" in subscription else False

			# "event" in the configuration can be a string, or
			# a list of strings.  If it's the former, convert it
			# into the latter.
			if not isinstance(events, (tuple, list, set)):
				events = [events]

			for event in events:
				if event not in self._subscriptions:
					self._subscriptions[event] = []
				self._subscriptions[event].append((command, commandType, debug))

				if event not in eventsToSubscribe:
					eventsToSubscribe.append(event)

		self.subscribe(eventsToSubscribe)

	def eventCallback(self, event, payload):
		"""
		Event callback, iterates over all subscribed commands for the given event, processes the command
		string and then executes the command via the abstract executeCommand method.
		"""

		GenericEventListener.eventCallback(self, event, payload)

		if not event in self._subscriptions:
			return

		for command, commandType, debug in self._subscriptions[event]:
			try:
				if isinstance(command, (tuple, list, set)):
					processedCommand = []
					for c in command:
						processedCommand.append(self._processCommand(c, event, payload))
				else:
					processedCommand = self._processCommand(command, event, payload)
				self.executeCommand(processedCommand, commandType, debug=debug)
			except KeyError:
				self._logger.warning("There was an error processing one or more placeholders in the following command: %s" % command)

	def executeCommand(self, command, commandType, debug=False):
		if commandType == "system":
			self._executeSystemCommand(command, debug=debug)
		elif commandType == "gcode":
			self._executeGcodeCommand(command, debug=debug)

	def _executeSystemCommand(self, command, debug=False):
		def commandExecutioner(cmd):
			if debug:
				self._logger.info("Executing system command: {}".format(cmd))
			else:
				self._logger.info("Executing a system command")
			# we run this with shell=True since we have to trust whatever
			# our admin configured as command and since we want to allow
			# shell-alike handling here...
			subprocess.check_call(cmd, shell=True)

		def process():
			try:
				if isinstance(command, (list, tuple, set)):
					for c in command:
						commandExecutioner(c)
				else:
					commandExecutioner(command)
			except subprocess.CalledProcessError as e:
				if debug:
					self._logger.warning("Command failed with return code {}: {}".format(e.returncode, str(e)))
				else:
					self._logger.warning("Command failed with return code {}, enable debug logging on target 'octoprint.events' for details".format(e.returncode))
			except Exception:
				self._logger.exception("Command failed")

		t = threading.Thread(target=process)
		t.daemon = True
		t.start()

	def _executeGcodeCommand(self, command, debug=False):
		commands = [command]
		if isinstance(command, (list, tuple, set)):
			commands = list(command)
		if debug:
			self._logger.info("Executing GCode commands: %r" % command)
		self._printer.commands(commands)

	def _processCommand(self, command, event, payload):
		"""
		Performs string substitutions in the command string based on a few current parameters.

		The following substitutions are currently supported:

		  - {__currentZ} : current Z position of the print head, or -1 if not available
		  - {__eventname} : the name of the event hook being triggered
		  - {__filename} : name of currently selected file, or "NO FILE" if no file is selected
		  - {__filepath} : path in origin location of currently selected file, or "NO FILE" if no file is selected
		  - {__fileorigin} : origin of currently selected file, or "NO FILE" if no file is selected
		  - {__progress} : current print progress in percent, 0 if no print is in progress
		  - {__data} : the string representation of the event's payload
		  - {__json} : the json representation of the event's payload, "{}" if there is no payload, "" if there was an error on serialization
		  - {__now} : ISO 8601 representation of the current date and time

		Additionally, the keys of the event's payload can also be used as placeholder.
		"""

		json_string = "{}"
		if payload:
			import json
			try:
				json_string = json.dumps(payload)
			except Exception:
				self._logger.exception("JSON: Cannot dump %r", payload)
				json_string = ""

		params = {
			"__currentZ": "-1",
			"__eventname": event,
			"__filename": "NO FILE",
			"__filepath": "NO PATH",
			"__progress": "0",
			"__data": str(payload),
			"__json": json_string,
			"__now": datetime.datetime.now().isoformat()
		}

		currentData = self._printer.get_current_data()

		if "currentZ" in currentData and currentData["currentZ"] is not None:
			params["__currentZ"] = str(currentData["currentZ"])

		if "job" in currentData and "file" in currentData["job"] and "name" in currentData["job"]["file"] \
				and currentData["job"]["file"]["name"] is not None:
			params["__filename"] = currentData["job"]["file"]["name"]
			params["__filepath"] = currentData["job"]["file"]["path"]
			params["__fileorigin"] = currentData["job"]["file"]["origin"]
			if "progress" in currentData and currentData["progress"] is not None \
				and "completion" in currentData["progress"] and currentData["progress"]["completion"] is not None:
				params["__progress"] = str(round(currentData["progress"]["completion"]))

		# now add the payload keys as well
		if isinstance(payload, dict):
			params.update(payload)

		return command.format(**params)
